using System;
using System.Collections.Generic;

using Amazon.ImportExport;
using Amazon.ImportExport.Model;
using Amazon.Runtime;

using Amazon.S3;
using Amazon.S3.Model;
using Amazon.S3.Util;
using CommonTests.Framework;
using NUnit.Framework;

namespace CommonTests.IntegrationTests
{
    [TestFixture]
    public class ImportExport : TestBase<AmazonImportExportClient>
    {
        const string IMPORT_MANIFEST =
@"bucket: @BUCKET@
accessKeyId: @ACCESS_KEY_ID@
manifestVersion: 1.3
eraseDevice: true
deviceId: 123
notificationEmail: john.doe@example.com
returnAddress:
   name: Amazon.com ATTN:Joe Random
   street1: 5555555 5th Ave
   city: Seattle
   stateOrProvince: WA
   postalCode: 98104
   phoneNumber: 206-555-1000
   country: USA
";

        const string EXPORT_MANIFEST =
@"manifestVersion: 1.2
accessKeyId: @ACCESS_KEY_ID@
deviceId: 532404500021
logBucket: @BUCKET@
trueCryptPassword: apassword
logPrefix: logs/
fileSystem: NTFS
notificationEmail: john.doe@example.com
operations:
    - exportBucket: @BUCKET@
returnAddress:
    name: Amazon.com ATTN Joe Random
    street1: 1200 12th Ave S.
    city: Seattle
    stateOrProvince: WA
    postalCode: 98114
    phoneNumber: 206-266-0000
    country: USA
";

        static AmazonS3Client s3Client;

        static string bucketName;

        [OneTimeSetUp]
        public void ClassInitialize()
        {
            s3Client = CreateClient<AmazonS3Client>();

            // Add test data to export
            bucketName = "sdk-import-test" + DateTime.Now.Ticks;
            s3Client.PutBucketAsync(new PutBucketRequest { BucketName = bucketName }).Wait();
            s3Client.PutObjectAsync(new PutObjectRequest
            {
                BucketName = bucketName,
                Key = "data.txt",
                ContentBody = "import-export-data"
            }).Wait();
        }

        [TearDown]
        public void Cleanup()
        {
            BaseClean();

            if(s3Client != null)
            {
                UtilityMethods.DeleteBucketWithObjectsAsync(s3Client, bucketName).Wait();
                s3Client.Dispose();
                s3Client = null;
            }
        }

        [Test]
        [Category("ImportExport")]
        public void TestImportExport()
        {
            // CreateJob
            CreateJobRequest createJobRequest = new CreateJobRequest
            {
                JobType = JobType.Import,
                Manifest = GetSampleManifestText(IMPORT_MANIFEST)
            };
            var createJobResponse = Client.CreateJobAsync(createJobRequest).Result;
            string createdJobId = createJobResponse.JobId;
            Assert.IsNotNull(createdJobId);
            Assert.AreEqual(JobType.Import, createJobResponse.JobType);
            Assert.IsNotNull(createJobResponse.Signature);
            Assert.IsNotNull(createJobResponse.SignatureFileContents);


            // UpdateJob
            UpdateJobRequest updateJobRequest = new UpdateJobRequest
            {
                JobId = createdJobId,
                JobType = JobType.Export,
                Manifest = GetSampleManifestText(EXPORT_MANIFEST)
            };
            Client.UpdateJobAsync(updateJobRequest).Wait();


            // ListJobs
            var listJobsResponse = Client.ListJobsAsync(new ListJobsRequest { MaxJobs = 100 }).Result;
            Assert.IsNotNull(listJobsResponse.IsTruncated);
            Job job = FindJob(createdJobId, listJobsResponse.Jobs);
            Assert.IsNotNull(job);
            Assert.IsTrue(job.CreationDate > DateTime.MinValue);
            Assert.IsFalse(job.IsCanceled);
            Assert.AreEqual(createdJobId, job.JobId);
            Assert.AreEqual(JobType.Export, job.JobType);
            Assert.IsFalse(job.IsCanceled);


            // GetStatus
            var getStatusResponse = Client.GetStatusAsync(new GetStatusRequest { JobId = createdJobId }).Result;
            Assert.IsNotNull(getStatusResponse.CreationDate);
            Assert.IsNotNull(getStatusResponse.CurrentManifest);
            Assert.AreEqual(createdJobId, getStatusResponse.JobId);
            Assert.AreEqual(JobType.Export, getStatusResponse.JobType);
            Assert.IsNotNull(getStatusResponse.ProgressMessage);
            Assert.IsNotNull(getStatusResponse.LocationMessage);
            Assert.IsNotNull(getStatusResponse.LocationCode);
            Assert.IsNotNull(getStatusResponse.Signature);
            Assert.AreEqual(0, getStatusResponse.ErrorCount);
            Assert.IsNotNull(getStatusResponse.ProgressMessage);
            Assert.IsNotNull(getStatusResponse.SignatureFileContents);
            Assert.IsNull(getStatusResponse.Carrier);
            Assert.IsNull(getStatusResponse.TrackingNumber);
            Assert.IsNull(getStatusResponse.LogBucket);
            Assert.IsNull(getStatusResponse.LogKey);


            // Cancel our test job
            Client.CancelJobAsync(new CancelJobRequest { JobId = createdJobId }).Wait();
            AssertJobIsCancelled(createdJobId);
            createdJobId = null;
        }

        private Job FindJob(String jobId, List<Job> jobs)
        {
            var job = jobs.Find(item => item.JobId == jobId);

            if (job == null)
                Assert.Fail("Expected to find a job with ID '" + jobId + "', but didn't");
            return job;
        }

        private void AssertJobIsCancelled(string jobId)
        {
            Job job = FindJob(jobId, Client.ListJobsAsync(new ListJobsRequest()).Result.Jobs);
            Assert.IsTrue(job.IsCanceled);
        }



        private string GetSampleManifestText(string manifest)
        {
            manifest = manifest.Replace("@BUCKET@", bucketName);
            manifest = manifest.Replace("@ACCESS_KEY_ID@", "AAAEXAMPLE");

            return manifest;
        }

    }
}
